#!/usr/bin/env python -i

# Skrypt pobiera napisy z logu OBS Studio - z wtyczką https://akabe.github.io/obs-transcript/
# filtruje powtórzenia, tworzy pośrednie wersje w plikach CSV. a na koniec wytwarza plik tT_0.SRT (wielowierszowe teksty, częste powtórzenia i poprawki)
# oraz tT_1.SRT - jednowierszowe, wyfiltrowane teksty. Numery w plikach CSV to numery pierwotnych próbek tekstu (ułatwia to porównywanie).
# Plik SRT można skopiować do foldera z filmem, nazwać go tak jak film z rozszerzeniem SRT i nanieść poprawki (w tym przesunięcia w czasie)
# np. programem https://www.nikse.dk/SubtitleEdit (gdzie można też wykonać automatyczne tłumaczenie na inny język).

# Przykład uruchomienia:
# py obs_tra.py "2021-06-25 11-26-55.txt"
# gdzie ostatni parametr to nazwa pliku logu obs w folderze "%AppData%\obs-studio\logs\"
# albo można wpisać nazwę pliku poniżej w wierszu in_Fname = ...
import re
import json
import csv
from datetime import datetime, time, timedelta
import os
import sys
appdata = os.getenv('APPDATA') # 'C:\\Users\\....\\AppData\\Roaming'
if len(sys.argv) > 1: in_FName = sys.argv[1]
else: 
  in_FName = "./2021-06-30 14-34-34x.txt" # "2021-06-24 14-52-10.txt" # "2021-06-25 11-26-55.txt" #

# Zdarza się powtórzenie "prawie" tego samego. Najwyższy wiersz bywa dość stabilny, 
# choć jest "migotanie" z wierszem wcześniejszym, najniższy jest ciągle poprawiany.
# Inne zjawisko to "zaleganie" ostatniego napisu, gdy jest cisza 
TIME_OUT = 2.5 #po tym czasie (sek.) napis "sam znika"
TIME_SILENT = 3 #wykrywanie milczenia
TIME_NEXT_TXT = 1.5 # kolejne wiersze podczas milczenia pojawiają się w tym rytmie
TIME_FILTER = 15 # czas filtrujący - gdy w tym czasie pojawia się ten sam napis, to jest ignorowany; to powinno być większe od TIME_OUT
TIME_SH = 0 # 5.8 #sek. dodatkowe przesunięcie podczas testowania - potem powinno być 0, a przesuwanie realizowane w edytorze np. Subtitle Edit

if not (('\\' in in_FName) or ('/' in in_FName)):
  in_FName = appdata + '\\obs-studio\\logs\\' + in_FName

def writeCsv(csvName, csvRows, csvHeader=None, br = None, encoding='utf-8-sig'):
  '''writeCsv(csvName: str, csvRows: [[],...], csvHeader: [] optional, br: str)
  br - replace '\n' with br
  '''
  with open(csvName,'w',newline='',encoding=encoding) as csvF:
    csvWr = csv.writer(csvF, delimiter=';',quoting=csv.QUOTE_MINIMAL)
    csvR = csvRows if csvHeader is None else [csvHeader] + csvRows
    if br is None:
      csvWr.writerows(csvR)
    else:
      csvWr.writerows([[str(s).replace('\n',br) for s in row] for row in csvR])

################# 0 #################

# 14:54:10.586: [obs-websocket] Request >> '{"source":"my_transcript_src","text":"Sprawdzam czy działa","request-type":"SetTextGDIPlusProperties","message-id":"19"}'
reTranscr = re.compile(r'''([\d:\.]{12}): \[obs-websocket\] Request >> '\{"source".+?,"text":(.+?),"request-type":''')

with open(in_FName, encoding='utf-8') as txt:
#  log = txt.read().splitlines()[:600]
  log = txt.read().splitlines()

time_text_0 = [] # odczytane [[time,text],[],...] z wierszy gdzie jest [obs-websocket] Request >>

for r in log:
  m = reTranscr.match(r)
  if m:
    timeLog = m[1]
    text = json.loads(m[2])
    time_text_0 += [[timeLog,text]]
    
# print(time_text_0)
writeCsv(csvName = "tT_0.csv", csvRows = time_text_0, csvHeader = ['time','text'], br='<br />')

################# 1 #################

(_number_, _time_, _text_) = range(3)

time_text_1 = []     # Lista: [nr, względny czas obliczany jak niżej, tekst]
time_text_az_1 = [] # Analogiczna lista  [czas obliczalny (naiwny, tu bez uwzględniania przejścia przez północ), uproszony tekst do porównania]
                    # do porównania - małe znaki, tylko znaki
time_start = None # pierwszy czas występujący w tabelach - do obliczania względnego czasu napisów

def to_az(s):
  return re.sub(r'[^\w]','',s.lower())

txs_prev_0 = ''
tme_prev = timedelta(seconds=0)
for i,  [tm, text] in enumerate(time_text_0):
  txs = text
  txs_az = '\n'.join([to_az(s) for s in txs.split('\n')]) # ; print('+',txs,'\n-',txs_az) #$#
  tme0 = datetime.strptime(tm+'000',"%H:%M:%S.%f")
  if not time_start: time_start = tme0                                             + timedelta(seconds=TIME_SH)
  tme = (tme0 - time_start) if tme0 > time_start else timedelta(seconds=0)
  time_text_1 += [[i, tme, txs]]
  time_text_az_1 += [[i, tme, txs_az]]

writeCsv(csvName = "tT_1.csv", csvRows = time_text_1,csvHeader = ['nr','time','text'], br='<br />')
writeCsv(csvName = "tT_1_az.csv", csvRows = time_text_az_1,csvHeader = ['nr','time','text'], br='<br />')
print('time_start:', time_start,)

# print('time_text_az_1 =================', time_text_az_1)
# >>>time_text_az_1 ================= [..., [1, datetime.timedelta(seconds=6, microseconds=554000), 'łałnodobrasprawdzamczydziała'], ...
#    gdy tekst pierwotny: "Łał-No dobra Sprawdzam czy działa\nrozpoznawanie\nmy"


################# 2 #################

# Porównujemy wg. time_text_az_1. Gdy dany tekst pojawi się po raz pierwszy, to zapamiętujemy jego czas 
# i ignorujemy kolejne wystąpienia tego tekstu przez np. 10 sek, 
# przy czym jako tekst wynikowy bierzemy ostatni z time_text_1 (jako najlepiej poprawiony, np. z przecinkami)

time_text_2 = [] # nr: [nr, tme, text]
time_text_az_2 = [] 
prev_time_text_az_2 = {} # text_az: [tme, txt] - słownik wykorzystanych ostatnio tekstów; tme - czas względem początku, txt - referencja do napisu org. w time_text_2
time_ignored = timedelta(seconds=TIME_FILTER) # przez taki czas pomijane będą duplikujące się napisy


for i, [nr, tme, tx_az] in enumerate(time_text_az_1):
  # print(nr, tme, tx_az) #>>> 1 0:00:06.554000 łałnodobrasprawdzamczydziała
  is_new =  tx_az not in prev_time_text_az_2
  if is_new or (tme > prev_time_text_az_2[tx_az][_time_] + time_ignored): # wstawienie danych klucza bo nie istniał lub się przeterminował
    #$# if not is_new: print('***',prev_time_text_az_2[tx_az][_time_],time_ignored,tme,prev_time_text_az_2[tx_az][_time_] + time_ignored < tme) #$#
    time_text_2 += [[nr, tme, time_text_1[i][_text_]]]
    prev_time_text_az_2[tx_az] = [nr, tme, time_text_2[-1]]
    time_text_az_2 += [[nr, tme, tx_az]] 
  else: #podstaw nowszy tekst oryginalny
    prev_time_text_az_2[tx_az][2][_text_] = time_text_1[i][_text_]

#$# print('\n'.join([f'''{nr}\t{tm}\t{tx}''' for nr, tm, tx in time_text_2])) #$#
writeCsv(csvName = "tT_2.csv", csvRows = time_text_2, csvHeader = ['nr','time','text'], br='<br />')
writeCsv(csvName = "tT_2_az.csv", csvRows = time_text_az_2,csvHeader = ['nr','time','text'], br='<br />')

################# 0.SRT #################

def timedeltaStr3(t):
  return (datetime(1900, 1, 1, 0)+t).strftime("%H:%M:%S,%f")[:-3]

def srt_info(nr, t1, t2, txt):
  return f'''{nr}\n{timedeltaStr3(t1)} --> {timedeltaStr3(t2)}\n{txt}\n\n'''

def writeSrt(srtName, data, secondsTimeOut=TIME_OUT):
  srt = ''
  nr = 0
  tme_prev = timedelta(seconds=0)
  for nr, tme, tx in data: #[[_number_,_time_,_text_],...]
    tme_max = tme_prev + timedelta(seconds=secondsTimeOut)
    tme_2 = tme if tme < tme_max else tme_max
    if nr > 0:
      srt += srt_info(nr, tme_prev, tme_2, tx_prev)
    tx_prev = tx
    tme_prev = tme
    nr += 1
  # jeszcze został ostatni wiersz do zapisania:
  srt += srt_info(nr, tme_prev, tme_2, tx_prev)
  with open(srtName,'w',newline='\r\n',encoding='utf-8-sig') as txt:
    txt.write(srt)


writeSrt(srtName='tT_0.srt', data=time_text_2, secondsTimeOut=TIME_OUT)

################# 3 #################

# Porównujemy wg. time_text_az_2 - pierwszy wiersz. Gdy dany tekst pojawi się po raz pierwszy, to zapamiętujemy jego czas 
# i ignorujemy kolejne wystąpienia tego tekstu przez np. 15 sek, 
# przy czym jako tekst wynikowy bierzemy ostatni pierwszy wiersz z time_text_2 (jako najlepiej poprawiony, np. z przecinkami)
# gdy nastąpi chwila milczenia to dokładamy kolejne wiersze po tym pierwszym.

time_text_3 = [ [0, timedelta(seconds=0), ''] ] # nr: [nr, tme, text]
prev_time_text_az_3 = {} # text_az: [tme, txt] - słownik wykorzystanych ostatnio tekstów; tme - czas względem początku, txt - referencja do napisu org. w time_text_3
time_ignored = timedelta(seconds=TIME_FILTER) # przez taki czas pomijane będą duplikujące się napisy

tme_prev = timedelta(seconds=0) # moment poprzedniego nowego tekstu - do wykrywania milczenia
txs_prev ,txs_az_prev = [''], ['']
for i, [nr, tme, tx_az] in enumerate(time_text_az_2):
  txs_az = tx_az.split('\n')
  tx_az0 = txs_az[0]
  txs = time_text_2[i][_text_].split('\n')
  tx0 = txs[0]
  
  #gdy wystąpiło milczenie to zrzuć kolejne wiersze
  if tme > tme_prev + timedelta(seconds=TIME_SILENT):
    for k in range(1,len(txs_az_prev)):
      tx_az1 = txs_az_prev[k]
      tx1 =  txs_prev[k]
      time_text_3 += [ [ nr, tme_prev + k*timedelta(seconds=TIME_NEXT_TXT), tx1 ] ]
      prev_time_text_az_3[ tx_az1 ] = [nr, tme, time_text_3[-1]]
    tme_prev = tme
    txs_az_prev = txs_az
    txs_prev = txs
  
  # nowe wpisy pierwszych wierszy
  is_new =  tx_az0 not in prev_time_text_az_3
  if is_new or (tme > prev_time_text_az_3[ tx_az0 ][_time_] + time_ignored): # wstawienie danych klucza bo nie istniał lub się przeterminował
    #$# if not is_new: print('***',tx_az0,'t_az',prev_time_text_az_3[ tx_az0 ][_time_],'tme:',tme,prev_time_text_az_3[ tx_az0 ][_time_] + time_ignored < tme) #$#
    #$# print('\n\n',str(prev_time_text_az_3).replace(']],',']],\n'),'\n') #$#
    time_text_3 += [ [ nr, tme, tx0 ] ]
    prev_time_text_az_3[ tx_az0 ] = [nr, tme, time_text_3[-1]]
    tme_prev = tme
    txs_az_prev = txs_az
    txs_prev = txs
  else: #podstaw nowszy (ulepszony) tekst oryginalny
    prev_time_text_az_3[ tx_az0 ][2][_text_] = tx0
    txs_az_prev = txs_az
    txs_prev = txs

print('\n'.join([f'''{nr}\t{tm}\t{tx}''' for nr, tm, tx in time_text_3]))
writeCsv(csvName = "tT_3.csv", csvRows = time_text_3, csvHeader = ['nr','time','text'], br='<br />')

################# 1.SRT #################

writeSrt(srtName='tT_1.srt', data=time_text_3, secondsTimeOut=TIME_OUT)
